;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "projection" "20260614.1822"
  "Project type support for `project'."
  '((emacs   "29.1")
    (project "0.9.8")
    (compat  "29.1.4.1")
    (f       "0.20")
    (s       "1.13"))
  :url "https://github.com/mohkale/projection"
  :commit "870a60b26416e13a7cd37ecfd863ae730d6a13fa"
  :revdesc "870a60b26416"
  :keywords '("project" "convenience")
  :authors '(("Mohsin Kaleem" . "mohkale@kisara.moe"))
  :maintainers '(("Mohsin Kaleem" . "mohkale@kisara.moe")))
